<?php

namespace Drupal\raven\EventSubscriber;

use Consolidation\AnnotatedCommand\CommandData;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Database\Database;
use Drupal\Core\Database\Event\StatementExecutionEndEvent;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Security\TrustedCallbackInterface;
use Drupal\raven\Logger\Raven;
use Sentry\SentrySdk;
use Sentry\Tracing\SpanContext;
use Sentry\Tracing\TransactionContext;
use Sentry\Tracing\TransactionSource;
use Symfony\Component\DependencyInjection\ContainerAwareInterface;
use Symfony\Component\DependencyInjection\ContainerAwareTrait;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Event\KernelEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Initializes Raven logger so Sentry functions can be called.
 */
class RequestSubscriber implements EventSubscriberInterface, ContainerAwareInterface, TrustedCallbackInterface {

  use ContainerAwareTrait;

  /**
   * Config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface|null
   */
  protected $configFactory;

  /**
   * {@inheritdoc}
   *
   * @var \Symfony\Component\DependencyInjection\ContainerInterface|null
   */
  protected $container;

  /**
   * The event dispatcher service.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * Raven logger service.
   *
   * @var \Drupal\raven\Logger\Raven|null
   */
  protected $logger;

  /**
   * Time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface|null
   */
  protected $time;

  /**
   * The transaction for HTTP requests.
   *
   * @var \Sentry\Tracing\Transaction|null
   */
  public $transaction;

  /**
   * Constructs the request subscriber.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory object.
   * @param \Drupal\raven\Logger\Raven $logger
   *   The logger service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher service.
   */
  public function __construct(ConfigFactoryInterface $config_factory = NULL, Raven $logger = NULL, TimeInterface $time = NULL, EventDispatcherInterface $event_dispatcher = NULL) {
    $this->configFactory = $config_factory;
    $this->logger = $logger;
    $this->time = $time;
    $this->eventDispatcher = $event_dispatcher;
  }

  /**
   * Sets up Drush command performance tracing.
   */
  public function onConsoleCommand(CommandData $commandData): void {
    $config = $this->configFactory->get('raven.settings');
    if (!$config->get('drush_tracing') || !$this->logger->getClient()) {
      return;
    }
    $transactionContext = new TransactionContext();
    $transactionContext->setName('drush ' . $commandData->input()->getArgument('command'));
    $transactionContext->setSource(TransactionSource::task());
    $transactionContext->setOp('drush.command');
    $transactionContext->setStartTimestamp($this->time->getRequestMicroTime());
    $this->transaction = \Sentry\startTransaction($transactionContext);
    SentrySdk::getCurrentHub()->setSpan($this->transaction);
    if ($config->get('database_tracing')) {
      foreach (Database::getAllConnectionInfo() as $key => $info) {
        Database::startLog('raven', $key);
      }
      // @todo The class_exists() check can be removed when support for Drupal
      // <10.1 is dropped.
      if (class_exists(StatementExecutionEndEvent::class)) {
        $this->eventDispatcher->addListener(StatementExecutionEndEvent::class, [
          static::class,
          'onStatementExecutionEnd',
        ]);
      }
    }
  }

  /**
   * Starts a transaction if performance tracing is enabled.
   *
   * @todo In Drupal 9+ the event should actually be RequestEvent.
   */
  public function onRequest(KernelEvent $event): void {
    $method = method_exists($event, 'isMainRequest') ? 'isMainRequest' : 'isMasterRequest';
    if (!$event->$method() || !$this->configFactory) {
      return;
    }
    $config = $this->configFactory->get('raven.settings');
    if (!$config->get('request_tracing') || !$this->logger || !$this->logger->getClient()) {
      return;
    }
    $request = $event->getRequest();
    $transactionContext = TransactionContext::fromHeaders($request->headers->get('sentry-trace', ''), $request->headers->get('baggage', ''));
    // This name will later be replaced with the route path, if possible.
    $transactionContext->setName($request->getMethod() . ' ' . $request->getUri());
    $transactionContext->setSource(TransactionSource::url());
    $transactionContext->setOp('http.server');
    $transactionContext->setTags([
      'http.method' => $request->getMethod(),
      'http.url' => $request->getUri(),
    ]);
    $transactionContext->setStartTimestamp($this->time->getRequestMicroTime());
    $this->transaction = \Sentry\startTransaction($transactionContext);
    SentrySdk::getCurrentHub()->setSpan($this->transaction);
    if ($config->get('database_tracing')) {
      foreach (Database::getAllConnectionInfo() as $key => $info) {
        Database::startLog('raven', $key);
      }
      // @todo The class_exists() check can be removed when support for Drupal
      // <10.1 is dropped.
      if (class_exists(StatementExecutionEndEvent::class)) {
        $this->eventDispatcher->addListener(StatementExecutionEndEvent::class, [
          static::class,
          'onStatementExecutionEnd',
        ]);
      }
    }
  }

  /**
   * Performance tracing.
   *
   * @todo In Drupal 9+ the event should actually be TerminateEvent.
   */
  public function onTerminate(KernelEvent $event): void {
    if (!$this->transaction) {
      return;
    }
    // Clean up the transaction name if we have a route path.
    if ($this->container->initialized('current_route_match')) {
      if ($route = $this->container->get('current_route_match')->getRouteObject()) {
        $this->transaction->setName($event->getRequest()->getMethod() . ' ' . $route->getPath());
        $this->transaction->getMetadata()->setSource(TransactionSource::route());
      }
    }
    $config = $this->configFactory->get('raven.settings');
    if (method_exists($event, 'getResponse')) {
      $statusCode = $event->getResponse()->getStatusCode();
      $this->transaction->setHttpStatus($statusCode);
      if ($statusCode === Response::HTTP_NOT_FOUND && !$config->get('404_tracing')) {
        $this->transaction->setSampled(FALSE);
      }
    }

    // @todo This code can be removed when support for Drupal <10.1 is dropped.
    if ($config->get('database_tracing') && !class_exists(StatementExecutionEndEvent::class)) {
      $this->collectDatabaseLog();
    }

    $this->transaction->finish();
  }

  /**
   * Create a span for each database statement.
   */
  public static function onStatementExecutionEnd(StatementExecutionEndEvent $event): void {
    $parent = SentrySdk::getCurrentHub()->getSpan();
    if (!$parent) {
      return;
    }
    $context = new SpanContext();
    $context->setOp('db.query');
    $context->setDescription($event->queryString);
    $context->setTags([
      'database' => $event->key,
      'target' => $event->target,
    ]);
    $context->setStartTimestamp($event->startTime);
    $context->setEndTimestamp($event->time);
    $parent->startChild($context);
  }

  /**
   * If database was initialized, create a span for each logged query.
   *
   * @todo This method will be removed when support for Drupal <10.1 is dropped.
   */
  public function collectDatabaseLog(): void {
    if (!$this->transaction || !$this->container || !$this->container->initialized('database')) {
      return;
    }
    $connections = [];
    foreach (Database::getAllConnectionInfo() as $key => $info) {
      try {
        $database = Database::getConnection('default', $key);
        if ($logger = $database->getLogger()) {
          $connections[$key] = $logger->get('raven');
        }
      }
      catch (\Exception $e) {
        // Could not connect.
      }
    }
    foreach ($connections as $key => $queries) {
      foreach ($queries as $query) {
        if (empty($query['start'])) {
          // Older versions of Drupal do not record query start time.
          return;
        }
        $context = new SpanContext();
        $context->setOp('db.query');
        $context->setDescription($query['query']);
        $context->setTags([
          'database' => $key,
          'target' => $query['target'],
        ]);
        $context->setStartTimestamp($query['start']);
        $context->setEndTimestamp($query['start'] + $query['time']);
        $this->transaction->startChild($context);
      }
    }
  }

  /**
   * Registers the methods in this class that should be listeners.
   *
   * @return mixed[]
   *   An array of event listener definitions.
   */
  public static function getSubscribedEvents() {
    $events[KernelEvents::REQUEST][] = ['onRequest', 222];
    $events[KernelEvents::TERMINATE][] = ['onTerminate', 222];
    return $events;
  }

  /**
   * {@inheritdoc}
   */
  public static function trustedCallbacks() {
    return ['getTraceparent'];
  }

  /**
   * Callback for returning the Sentry trace string as renderable array.
   *
   * @return mixed[]
   *   Renderable array.
   */
  public function getTraceparent(): array {
    $markup = '';
    if (class_exists(SentrySdk::class)) {
      if ($span = SentrySdk::getCurrentHub()->getSpan()) {
        $markup = $span->toTraceparent();
      }
    }
    return ['#markup' => $markup];
  }

}
